import mx.central.data.LCService;
import mx.central.data.LCDataProvider;
import mx.central.services.WebService;

/**
 * StockService class
 * Provides the API methods for the StockAgent, manages SharedObject storage,
 * and handles communications with the web service.
 */
class StockService implements IStockServer {

    var oStorage:SharedObject;  // local shared object holding all
                                // stock and interface data
    var oStockClient;           // local connection to communication
                                // with apps and pods

    var oCompanies:LCDataProvider; // server side of data, connected
    var oRules:LCDataProvider;  //    to oStorage objects
    
    var bOnline:Boolean = true;           // online/offline state
    var oStockWebService:WebService; // used to access stock data

    var oAgent:StockAgent;      // handy for calling back into the
                                // parent object to create notices, etc

    // enum for different alert actions
    static var kNoCompany = -1;

    function StockService(agent:StockAgent)
    {
		StockUtils.trace("constructor", "StockService");
        oAgent = agent;
        
        // begin local communication service between this agent and
        // apps/pods
        oStockClient = LCService.createServer(StockInterfaces, this, false);
        oStockClient.allowDomain = function(domain:String):Boolean {
			return true;
		};	
		//StockUtils.traceProps("oStockClient=",oStockClient);
		
        oCompanies = LCDataProvider.createServer("companies");
		//StockUtils.traceProps("oCompanies=",gCompanies);
		
        oRules = LCDataProvider.createServer("rules");
		//StockUtils.traceProps("oRules=",gRules);
	
        initStorage();
        
        // get initial data from local cache
        oCompanies.setData(oStorage.data.companies);
        oRules.setData(oStorage.data.rules);

        // get reference to stock data web service
        oStockWebService = new WebService("http://www.flash-db.com/services/ws/companyInfo.wsdl");
    }
	
	public function allowDomain(domain:String):Boolean {
		return true;
	}
	
	/**
	 * Called if the LCService connection fails
	 */
	public function createServer_Status (clientObject:Object) {
		StockUtils.trace("LCService.createServer failed.");
	}
	
	/**
	 * Called when the LCService connection to the AIM service is successful
	 */
	function createServer_Result (proxyObject:Object) {
		StockUtils.trace("LCService.createServer succeeded.");
	}

    //-------------------------------------------------------------------
    // Last Update Date
    function getLastUpdate():Date
    {
        return oStorage.data.lastUpdate;
    }

    //-------------------------------------------------------------------
    // Companies

    function stockSearch(symbol:String):StockItem
    {
		StockUtils.trace("stockSearch(" + symbol + ")", "StockService");
		
        // stockSearch does its best to return information about a company:
        // either find it in local cache or call web service if it's a new one
        // we return an error if we don't have it and are offline
		if (!haveCompany(symbol)) {
            // we don't have a local copy of this stock symbol
            if (bOnline) {
                // we're online, so request it
                addCompany(symbol);
            } else {
                // we're offline, so return an error message
                return(new StockItem(symbol, "(unavailable offline)"));
            }			
		}

		return getCompany(symbol);
    }

    function getCompany(symbol:String):StockItem
    {
		StockUtils.trace("getCompany(" + symbol + ")", "StockService");
        var item:StockItem;
        var i:Number = findCompany(symbol);
        if (i != kNoCompany) {
			var obj:Object = oCompanies.getItemAt(i);
            item = StockItem.create(obj);
        }
        return item;
    }

    function haveCompany(symbol:String):Boolean
    {
        return (findCompany(symbol) != kNoCompany);
    }
    
    /**
	 * Returns a unique index for the given symbol
	 */
    function findCompany(symbol):Number
    {
		StockUtils.trace("findCompany(" + symbol + ")", "StockService");
        for (var i:Number=0; i<oStorage.data.companies.length; i++) {
            
            if (oStorage.data.companies[i].symbol.toUpperCase() ==
                symbol.toUpperCase())
			{
                return i;
			}
        }
        return kNoCompany;
    }

    function addCompany(symbol:String):Void
    {
		StockUtils.trace("addCompany(" + symbol + ")", "StockService");
		
        // if not already added, add it
        if (!haveCompany(symbol)) {
			StockUtils.trace(" before companies.length=" + oStorage.data.companies.length, "StockService" );
			StockUtils.trace(" last symbol=" + oStorage.data.companies[oStorage.data.companies.length-1].symbol, "StockService" );
			
            oStorage.data.companies.push(new StockItem(symbol));
            oStorage.flush();
			
			StockUtils.trace(" after companies.length=" + oStorage.data.companies.length, "StockService" );
			StockUtils.trace(" last symbol=" + oStorage.data.companies[oStorage.data.companies.length-1].symbol, "StockService" );

            // look up its price
            updateCompany(symbol);
        }
    }

    function removeCompany(symbol:String):Void
    {
		StockUtils.trace("removeCompany(" + symbol + ")", "StockService");
        // get the current selection
        var i:Number = findCompany(symbol);
	
        // remove it
        oStorage.data.companies.splice(i,1);
        oStorage.flush();
	
        // tell the Stock app to redraw its view
        broadcastCompanies();
    }

	/**
	 * 
	 */
    function updateCompany(symbol:String):Void
    {
		StockUtils.trace("updateCompany(" + symbol + ")", "StockService");
        // scope workaround
        var ss:StockService = this;
        
        var stockRequest =
            oStockWebService.doCompanyInfo("anyuser", "anypassword", symbol);
        
        // handle the result when it returns
        stockRequest.onResult = function(result) {
			StockUtils.trace("stockRequest.onResult()", "StockService");
            ss.saveStock(result);
        }

        stockRequest.onFault = function(fault) {
			StockUtils.trace("stockRequest.onFault()", "StockService");
			StockUtils.traceProps("fault=", fault);
            // handle faults here!
        }
        
    }

	/**
	 * 
	 */
    function saveStock(result:StockItem):Void
    {
		StockUtils.trace("saveStock(" + result.symbol + ")", "StockService");
        // figure out which stock the result is for
        var i:Number = findCompany(result.symbol);
        if (i != kNoCompany) {
            // cache data
            oStorage.data.lastUpdate = new Date();
            oStorage.data.companies[i] = result;
            oStorage.flush();

            // process alert rules on result
            processRules(result);
        }
	
        // tell the apps and pods about the new info
        broadcastCompanies();
        oStockClient.setCompany(result);
    }

    function broadcastCompanies()
    {
		StockUtils.trace("broadcastCompanies()", "StockService");
        oStockClient.setLastUpdate(oStorage.data.lastUpdate);
        oCompanies.removeAll();
        oCompanies.setData(oStorage.data.companies);
    }

    function setLastUpdate(lastUpdate) {
		StockUtils.trace("setLastUpdate(" + lastUpdate + ")", "StockService");
        oStockClient.setLastUpdate(lastUpdate);
    }

    //-------------------------------------------------------------------
    // Rules

    function broadcastRules(Void):Void
    {
		StockUtils.trace("broadcastRules()", "StockService");
        oRules.removeAll();
        oRules.setData(oStorage.data.rules);
    }

    function findRule(ruleID):Number
    {
        for (var i:Number=0; i<oStorage.data.rules.length; i++) {
            if (oStorage.data.rules[i].id == ruleID)
                return i;
        }
        
        return kNoCompany;
    }

    function addRule(rule:AlertRuleItem):Number
    {
		StockUtils.trace("addRule(" + rule.symbol + ")", "StockService");
        // assign unique id
        rule.id = oStorage.data.nextRuleID++;
		
        rule.lastDismissedTime = 0;
        rule.lastExtreme = rule.value;
        rule.lastExtremeTime = 0;
        rule.lastDisplayTime = 0;
        rule.lastResult = 0;
        rule.noticeID = 0;
	
        oStorage.data.rules.push(rule);
        oStorage.flush();
        broadcastRules();
	
        return(rule.id);
    }

    function removeRule(ruleID:Number):Void
    {
		StockUtils.trace("removeRule(" + ruleID + ")", "StockService");
        var i:Number = findRule(ruleID);

        oStorage.data.rules.splice(i,1);
        oStorage.flush();
        broadcastRules();
    }

    function saveRule(rule:AlertRuleItem):Void
    {
		StockUtils.trace("saveRule(" + rule.id + ")", "StockService");
		//StockUtils.traceProps("rule=", rule);
        
		if (rule.id != undefined) {
			// reset last interaction if rule is changed
			rule.lastDismissedTime = 0;
			rule.lastExtreme = rule.value;
			rule.lastExtremeTime = 0;
		
			var i:Number = findRule(rule.id);
			oStorage.data.rules[i] = rule;
			oStorage.flush();
			broadcastRules();
		}
    }

    function think(Void):Boolean
    {
		StockUtils.trace("think()", "StockService");
		
        // request updated information for each company in our list
        for (var i:Number=0; i<oStorage.data.companies.length; i++) {
            updateCompany(oStorage.data.companies[i].symbol);
        }
	
        // refresh the update date/time in the apps and pods
        setLastUpdate(oStorage.data.lastUpdate);
		
		// return a value so we know to hide the progress bar when done
		return true;
    }


    function initStorage()
    {
        oStorage = SharedObject.getLocal("stockwatcher1.5b");
	
        // DEBUG: uncomment the lines below to reset information every time we launch
        // oStorage.data.companies = undefined;
        // oStorage.data.rules = undefined;
        // oStorage.data.testvalues = undefined;
	
        if (oStorage.data.companies == undefined)
		{
			oStorage.data.companies = new Array();
		}
		StockUtils.trace("oStorage companies.length=" + oStorage.data.companies.length, "StockService" );
		StockUtils.trace(" last symbol=" + oStorage.data.companies[oStorage.data.companies.length-1].symbol, "StockService" );

        if (oStorage.data.rules == undefined)
		{
			oStorage.data.rules = new Array();
			oStorage.data.nextRuleID = 1;
		}
		StockUtils.trace("oStorage rules.length=" + oStorage.data.rules.length, "StockService" );
		StockUtils.trace(" last symbol=" + oStorage.data.rules[oStorage.data.rules.length-1].symbol, "StockService" );
	
        if (oStorage.data.testvalues == undefined)
		{
			oStorage.data.testvalues = new Array();
		}
    }
	/**
	 * Called when uninstalling this application, in order to clear out the shared objects
	 */
	function deleteStorage(Void):Void
    {
		// purges all data from the SharedObject and deletes it from disk
        oStorage.clear();
    }

    function processNoticeEvent(noticeData:Object) {
        
        // find rule that corresponds to this notice, if any
        for (var i:Number=0; i<oStorage.data.rules.length; i++) {

            if (oStorage.data.rules[i].noticeID == noticeData.id) {
                // TODO check to see if was actually dismissed or just opened
                // clear this reference to the notice
                oStorage.data.rules[i].lastDismissedTime = new Date();
                oStorage.data.rules[i].noticeID = 0;
            }
        }
    }

    function processRules(result)
    {
        /* Alert processing:
	
		if rule matches
        - if dismissed already today and if new extreme match then create notice
        - if not shown yet today, create one
        - if existing notice, update with extreme of match (highest high, lowest low)
		
		if existing notice about symbol
        - update with current value with time stamp
        */

        var now:Date = new Date();
	
        // step through each rule
        for (var i:Number=0; i<oStorage.data.rules.length; i++) {
            var rule = oStorage.data.rules[i];
            
            // if this rule matches the current symbol, process it
            if (rule.symbol.toUpperCase() == result.symbol.toUpperCase()) {
                rule.lastResult = result[rule.criteria];
                
                var shouldAlert:Boolean = (rule.alert == AlertRuleItem.kBringToAttention);
                var showNotice:Boolean = false;
				
                // if last extreme time is over a day old, reset it
                if (((now.getTime() - rule.lastDisplayTime.getTime()) > 86400000) || 
                    (now.getDay() != rule.lastDisplayTime.getDay())) {
                    
                    rule.lastExtreme = rule.value;
                }
				
                // check for triggering rule: if result exceeds the
                // lastExtreme (which starts as the rule value)
                if (((rule.compare == "over") &&
                     (Number(rule.lastResult) > Number(rule.lastExtreme))) ||
                    ((rule.compare == "under") &&
                     (Number(rule.lastResult) < Number(rule.lastExtreme)))) {
                    
                    rule.lastExtreme = rule.lastResult;
                    rule.lastExtremeTime = result.tradedTime;
                    showNotice = true;
                }
				
                // if there is a notice already displayed for this
                // rule, be polite and just update it
                if (rule.noticeID > 0) {

                    // remove current notice, will add it back below
                    oAgent.removeNotice(rule.noticeID);
                    rule.noticeID = 0;
                    showNotice = true;
                    shouldAlert = false; // don't alert when updating existing notice
                }
                
                if (showNotice) {

                    /* Create notice in format:
                     * MACR 14.5
                     *   						
                     * 18.75 high 3:00pm
                     * 14.5 most recent 3:50pm
                     *
                     * Rule: notice when lastTrade > 15
                     */
                    var noticeName:String =
                        rule.symbol.toUpperCase() + " " + rule.lastResult;
                        
                    var description:String =
                        rule.lastExtreme + " " +
                        (rule.compare == "over" ? "high" : "low") + " " +
                        rule.lastExtremeTime;
                        
                    if ((rule.lastResult != 0) && (rule.lastExtreme != rule.lastResult)) {

                        description += "\r" + rule.lastResult +
                            " recent " + rule.result.tradedTime;
                    }
                        
                    description += "\r\rRule:\rnotice when " + rule.criteria +
                        (rule.compare == "over" ? " > " : " < ") + rule.value;
                        					
                    rule.lastDisplayTime = now;
                    rule.noticeID =
                        oAgent.postNotice(rule.symbol, noticeName, description, shouldAlert);
                }
            }
        }
        oStorage.flush();
    }
    
    
    //-------------------------------------------------------------------
    // Debug

    // set fake stock values to test notices
    function debugSetStock(ticker)
    {
        if (ticker.symbol == "free") {
            delete oStorage.data.testvalues;
            oStorage.data.testvalues = new Array();
            
            // reset last seen extreme values
            for (var i:Number=0; i<oStorage.data.rules.length; i++)
                {
                    oStorage.data.rules[i].lastExtreme = oStorage.data.rules[i].value;
                }
        }
        else {
            oStorage.data.testvalues[ticker.symbol] = ticker.value;
        }
    }
}
